﻿// Copyright 2009 Oliver Sturm <oliver@sturmnet.org> All rights reserved. 

#light

open System

open Helpers

// a recursive function
let rec fact x = if x = 0 then 1 else x * fact (x - 1)

// a lambda expression - signature is the same for these two functions
let add = fun x y -> x + y
let mult x y = x * y

Output.banner "Functions as first-class citizens"

let outputFiltered f l = 
    for e in l do
        if f e then
            printf "%A " e
    printfn ""
    
// passing functions around    
let outputEvens list = 
    let isEven x = x % 2 = 0
    outputFiltered isEven list
    // or:
    // outputFiltered (fun x -> x % 2 = 0) lest

outputEvens [3;6;9;12;15;18;21]

// ------------- snip (demo)


Output.banner "Nested calls, piping and composition"

// nesting calls
printfn "%d" (mult (add 12 3) (mult 12 (add 17 4)))

// chaining (piping) calls
printfn "%d" (add 17 4 |> mult 12 |> mult (add 12 3))

// composition
let mult3add7 = mult 3 >> add 7
printfn "%d" (mult3add7 12)

// ------------- snip (demo)

// closures

Output.banner "Closures"

let testFunction = 
    let x = 20
    let addSomething v = x + v
    printfn "%d" (addSomething 15)
    addSomething
    
let outerAddSomething = testFunction
printfn "%d" (outerAddSomething 35)

// ------------- snip (demo)


// curried function, separately
let addSeparate x = 
    fun y -> x + y
    
// partial application calls the first function, 
// remembering the second
let add5separate = addSeparate 5

// F# curries automatically, so same result as above
// We've already got this:
// let add x y = x + y
let add5 = add 5

Output.banner "Precomputation"

// The basic implementation of this function creates
// an optimized lookup data structure, but since it happens
// every single time, there's no benefit
let isInListDumb (list: 'a list) v = 
    let lookupTable = new HashSet<'a>(list)
    printfn "Lookup table created, looking up value"
    lookupTable.Contains v
    
printfn "%b" (isInListDumb [10;20;32;47] 20)
printfn "%b" (isInListDumb [10;20;32;47] 30)
printfn "%b" (isInListDumb [10;20;32;47] 47)

// It seems a good idea to use partial application
// in order to create the lookup data structure just once
// Doesn't work like that, though...
let cleverTry1 = isInListDumb [10;20;32;47]
printfn "%b" (cleverTry1 20)
printfn "%b" (cleverTry1 30)
printfn "%b" (cleverTry1 47)

// Creating a function with "manual" currying allows 
// inserting additional functionality in the sequence
let isInListClever (list: 'a list) = 
    let lookupTable = new HashSet<'a>(list)
    printfn "Clever lookup table created"
    fun v ->
        printfn "looking up value" 
        lookupTable.Contains v

// This time the function can use partial application
// and the lookup data structure is created just once
let cleverTry2 = isInListClever [10;20;32;47]
printfn "%b" (cleverTry2 20)
printfn "%b" (cleverTry2 30)
printfn "%b" (cleverTry2 47)


// is x a prime number or not...
let isPrime x =
    printfn "Calculating prime-ness of %d" x
    let mutable result = true 
    for div in 2..(x/2) do
        if x % div = 0 then
            result <- false
    result
    
Output.banner "Memoization"
    
open System.Collections.Generic

// a result cache
let cache = new Dictionary<int, bool>()

// return cached result if possible, otherwise calculate
// and cache
let cachedIsPrime x = 
    if not (cache.ContainsKey(x)) then
        cache.Add(x, isPrime(x))
    cache.[x]

// A little helper function for the output
let outputResult x f =
    printfn "Primeness of %d is %b" x (f x)

// simple call calculates prime-ness every time, of course
outputResult 7 isPrime
outputResult 7 isPrime

// using the cached version, prime-ness gets calculated 
// just once (per input value)
outputResult 7 cachedIsPrime
outputResult 7 cachedIsPrime


// Memoization implements the above caching pattern in a function
let memoize (f: 'a -> 'b) =
    let cache = new Dictionary<'a, 'b>()
    fun v ->
        if not (cache.ContainsKey(v)) then 
            cache.Add(v, f(v))
        cache.[v]

// generate a memoized version of isPrime
let isPrimeMem = memoize isPrime

outputResult 7 isPrimeMem
outputResult 7 isPrimeMem

// memoizing functions with multiple parameters isn't easy
// this is the "manual" way
let addMemManual = 
    memoize (fun x -> 
        memoize (fun y -> 
            printfn "Adding %d to %d" x y 
            x + y))

printfn "%d" (addMemManual 10 20)
printfn "%d" (addMemManual 10 20)

// Would be nicer to memoize the function all the way
// without having to change the implementation of the 
// function that is being memoized.
// This is a rather imperative implementation.
open System.Reflection

let deepMemoize (f: 'a -> 'b) =
    let cache = new Dictionary<'a, 'b>()

    fun v ->
        if not (cache.ContainsKey(v)) then
            printfn "Memoizing result for %A" v
            let result = f(v);
            let resultType = typeof<'b>
            if resultType.IsGenericType && 
                resultType.GetGenericTypeDefinition() = typedefof<FastFunc<_,_>> then
                let deepMemoizeFunction = 
                    Type.GetType("Module2").GetMethod("deepMemoize", 
                        BindingFlags.Public ||| BindingFlags.Static)
                let typedDeepMemoizeFunction = 
                    deepMemoizeFunction.MakeGenericMethod(resultType.GetGenericArguments())
                let memoizedResult = 
                    typedDeepMemoizeFunction.Invoke(null, [| (box result) |]) :?> 'b
                cache.Add(v, memoizedResult)
            else
                cache.Add(v, result)
        else
            printfn "Returning result for %A from memory" v            

        cache.[v]

let addMem = deepMemoize add

printfn "%d" (addMem 10 20)
printfn "%d" (addMem 10 20)


